using System;
using System.Collections.Generic;

namespace MMBot
{
    public class IRCChannel
    {
        public string Name { get; set; }
        public string Keyword { get; set; }
        public bool Hidden { get; set; }
        public string Topic { get; set; }
        [Newtonsoft.Json.JsonIgnore]
        public List<IRCUser> People = new List<IRCUser>();
        public List<IRCChanUserStats> stats = new List<IRCChanUserStats>();
        [Newtonsoft.Json.JsonIgnore]
        public System.Timers.Timer CommandTimer = new System.Timers.Timer(5000)
        {
            AutoReset = false,
            Enabled = false
        };
        [Newtonsoft.Json.JsonIgnore]
        public System.Timers.Timer WaitTimer = new System.Timers.Timer(10000)
        {
            AutoReset = false,
            Enabled = false
        };
        public bool greeting;
        public int Consecutivecmds;
        public bool Active = true;
        public LinkCheckMode linkcheck = LinkCheckMode.Off;
        public List<string> akick = new List<string>();
        public List<string> VOP = new List<string>();
        public List<string> HOP = new List<string>();
        public List<string> AOP = new List<string>();
        public List<Feed> feeds = new List<Feed>();
        public ulong peakusers;
        public bool displayaccesserror = true;
        public Dictionary<string, UserModes> AccessLevels = new Dictionary<string, UserModes>();
        public List<Tuple<string, string, DateTime, string>> Notes = new List<Tuple<string, string, DateTime, string>>();

        [Newtonsoft.Json.JsonIgnore]
        public IRC IrcObject;

        public IRCChannel()
        {
            CommandTimer.Elapsed += CMD_Tick;
            WaitTimer.Elapsed += Wait_Tick;
        }

        public IRCChannel(IRC network, string ChannelName)
            : this()
        {
            IrcObject = network;
            Name = ChannelName;
            Keyword = string.Empty;
        }

        public IRCChannel(IRC network, string ChannelName, string ChannelKeyword)
            : this(network, ChannelName)
        {
            Keyword = ChannelKeyword;
        }

        internal void CMD_Tick(object sender, System.Timers.ElapsedEventArgs e)
        {
            CommandTimer.Stop();
            Consecutivecmds = 0;
        }

        internal void Wait_Tick(object sender, System.Timers.ElapsedEventArgs e)
        {
            WaitTimer.Stop();
            IrcObject.WriteMessage("Commands reenabled.", Name);
        }

        public IRCUser GetUser(string name)
        {
            foreach (IRCUser user in People)
            {
                if (user.name.Equals(name, StringComparison.CurrentCultureIgnoreCase))
                    return user;
            }
            return null;
        }

        public IRCUser[] GetUsersByMask(string mask)
        {
            List<IRCUser> result = new List<IRCUser>();
            foreach (IRCUser item in People)
                if (item.Mask.Like(mask))
                    result.Add(item);
            return result.ToArray();
        }

        public IRCChanUserStats GetUserStats(string name)
        {
            List<string> a = IrcObject.GetAliases(name);
            if (a == null)
                a = new List<string>() { name };
            foreach (IRCChanUserStats s in stats)
                foreach (string al in a)
                    if (al.Equals(s.name, StringComparison.OrdinalIgnoreCase))
                        return s;
            return null;
        }

        public IRCChanUserStats GetUserStats(IRCUser person)
        {
            List<string> a = IrcObject.GetAliases(person.name);
            foreach (IRCChanUserStats s in stats)
                foreach (string al in a)
                    if (al.Equals(s.name, StringComparison.OrdinalIgnoreCase))
                        return s;
            return null;
        }

        public void AddUser(string name)
        {
            if (GetUser(name.TrimStart(IrcObject.prefixes)) != null) return;
            People.Add(new IRCUser(name, IrcObject));
            peakusers = Math.Max(peakusers, (ulong)People.Count);
            if (GetUserStats(name.TrimStart(IrcObject.prefixes)) == null)
                stats.Add(new IRCChanUserStats(name.TrimStart(IrcObject.prefixes)));
            else
            {
                IRCChanUserStats stat = GetUserStats(name.TrimStart(IrcObject.prefixes));
                stat.name = name.TrimStart(IrcObject.prefixes);
                stat.refcount++;
                if (!stat.onlinetimer.IsRunning)
                    stat.onlinetimer.Start();
            }
            if (IrcObject.GetAliases(name.TrimStart(IrcObject.prefixes)) == null)
            {
                List<string> x = new List<string>();
                x.Add(name.TrimStart(IrcObject.prefixes));
                IrcObject.Aliases.Add(x);
            }
        }

        public void RemoveUser(string name)
        {
            foreach (IRCUser user in People)
            {
                if (user.name.Equals(name, StringComparison.CurrentCultureIgnoreCase))
                {
                    People.Remove(user);
                    return;
                }
            }
        }

        public class Feed
        {
            public string url;
            public string title;
            public System.DateTime lastupdate;

            public Feed() { }
            public Feed(string url, DateTime lastupdate)
            {
                this.url = url;
                this.lastupdate = lastupdate;
            }

            public Feed(string url)
            {
                this.url = url;
                for (int eqq = 1; eqq <= 5; eqq++)
                {
                    try
                    {
                        System.Xml.XmlDocument x = new System.Xml.XmlDocument();
                        System.Net.HttpWebRequest s = (System.Net.HttpWebRequest)System.Net.HttpWebRequest.Create(url);
                        s.Credentials = System.Net.CredentialCache.DefaultCredentials;
                        s.UserAgent = Module1.useragent;
                        s.Timeout = 10000;
                        System.Net.HttpWebResponse response = (System.Net.HttpWebResponse)s.GetResponse();
                        System.IO.Stream dataStream = response.GetResponseStream();
                        System.IO.StreamReader reader = new System.IO.StreamReader(dataStream);
                        string responseFromServer = reader.ReadToEnd();
                        x.LoadXml(responseFromServer);
                        System.Xml.XmlNode node = x["rss"];
                        if (node == null)
                        {
                            node = x["feed"];
                            title = node["title"].InnerText;
                            System.Xml.XmlNodeList nodes = node.ChildNodes;
                            for (int i = nodes.Count - 1; i >= 0; i += -1)
                            {
                                node = nodes[i];
                                if (!node.Name.Equals("entry", StringComparison.CurrentCultureIgnoreCase))
                                    continue;
                                System.DateTime update = default(System.DateTime);
                                if (node["updated"] != null)
                                {
                                    update = System.DateTime.Parse(node["updated"].InnerText);
                                }
                                else
                                {
                                    update = System.DateTime.Parse(node["created"].InnerText);
                                }
                                if (lastupdate < update)
                                    lastupdate = update;
                            }
                        }
                        else
                        {
                            node = node["channel"];
                            title = node["title"].InnerText;
                            System.Xml.XmlNodeList nodes = node.ChildNodes;
                            for (int i = nodes.Count - 1; i >= 0; i += -1)
                            {
                                node = nodes[i];
                                if (!node.Name.Equals("item", StringComparison.CurrentCultureIgnoreCase))
                                    continue;
                                System.DateTime update = System.DateTime.Parse(node["pubDate"].InnerText);
                                if (lastupdate < update)
                                    lastupdate = update;
                            }
                        }
                        break;
                    }
                    catch
                    {
                        //IrcObject.WriteMessage("Feed error: " & Feed.url & " " & ex.GetType().Name & ": " & ex.Message, OpName)
                        //IrcObject.WriteMessage(ex.StackTrace, OpName)
                    }
                    System.Threading.Thread.Sleep(100);
                }
                if (lastupdate == System.DateTime.MinValue)
                    lastupdate = DateTime.Now;
            }
        }
    }

    public enum LinkCheckMode
    {
        Off,
        TitleOnly,
        On
    }
}